// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;

/**
 *  This class encapsulates a single trap placed on a given node. The
 *  traps for a given property name on a given box are maintained as a
 *  linked list stack, with the most recently placed trap at the head
 *  of the list.
 */
public class Trap {

    // Static Data //////////////////////////////////////////////////////////////

    /** List of properties that cannot be trapped */
    private static final Hash PROHIBITED = new Hash(120, 3);

    static {
        String[] p = new String[] {
            "shrink", "hshrink", "vshrink", "x", "y",
            "width", "height", "flex", "colspan", "rowspan", "cols",
            "rows", "align", "invisible", "absolute", "globalx",
            "globaly", "minwidth", "minheight", "height", "width",
            "maxwidth", "maxheight", "numchildren", "hpad", "vpad",
            "buffered", "cursor", "mousex", "mousey",
            "mouseinside", "thisbox", "indexof", "path", "font", "fontsize"
        };
        for(int i=0; i<p.length; i++) PROHIBITED.put(p[i], Boolean.TRUE);
    };


    // Instance Members ////////////////////////////////////////////////////////

    /** the box on which this trap was placed */
    private Box trapee = null;

    /** the function for this trap */
    JS.CompiledFunction f = null;

    /** the next trap down the trap stack */
    private Trap next = null;

    /** the property that the trap was placed on */
    private Object name = null;


    // Static Methods //////////////////////////////////////////////////////////////////////////

    /**
     *  adds a trap.
     *  @param trapee the box to place the trap on
     *  @param name the name of the property to trap on
     *  @param f the function to place as a trap
     */
    static void addTrap(Box trapee, Object name, JS.CompiledFunction f) {

        // check if this script has already placed a trap on this property
        for(Trap t = (Trap)trapee.get(name, Trap.class); t != null; t = t.next)
            if (t.f == f) return;
        
        // actually place the trap
        Trap t = new Trap();
        t.next = (Trap)trapee.get(name, Trap.class);
        trapee.put(name, Trap.class, t);
        t.trapee = trapee;
        t.name = name;
        t.f = f;
    }


    /**
     *  deletes a trap.
     *  @param trapee the box to remove the trap from
     *  @param name the name of the property to trap on
     *  @param f the function to remove
     */
    static void delTrap(Box trapee, Object name, JS.CompiledFunction f) {
        Trap t = (Trap)trapee.get(name, Trap.class);
        if (t.f == f) {
            trapee.put(name, Trap.class, t.next);
            return;
        }
        for(; t.next != null; t = t.next)
            if (t.next.f == f) {
                t.next = t.next.next;
                return;
            }
        Log.logJS("warning: tried to remove a trap that had not been placed");
    }


    // Instance Methods //////////////////////////////////////////////////////////////////////////

    private Trap() { }

    public Object perform() throws JS.Exn {
        try {
            if (f.getNumFormalArgs() > 0) return cascade();
            return f.call(new TrapArgs(this));
        } catch (Exception e) {
            Log.log(this, "Exception thrown from within trap: " + e);
            return null;
        }
    }
    
    public void perform(Object val) throws JS.Exn {
        try {
            if (f.getNumFormalArgs() == 0) cascade(val);
            TrapArgs ta = new TrapArgs(this, val);
            Object ret = f.call(ta);
            if (ret != Boolean.FALSE && !ta.cascadeHappened) cascade(val);
        } catch (Exception e) {
            Log.log(this, "Exception thrown from within trap: " + e);
        }
    }
    
    public Object cascade() {
        if (next != null) return next.perform();
        return trapee.get(name, true);
    }

    public void cascade(Object val) {
        if (next != null) next.perform(val);
        trapee.put(name, val, true);
    }

    private static class TrapArgs extends JS.Array {
        private Trap t;
        public boolean cascadeHappened = false;
        public TrapArgs(Trap t) { this.t = t; }
        public TrapArgs(Trap t, Object value) { this.t = t; addElement(value); }
        
        public void put(Object key, Object val) {
            if (key.equals("cascade")) { cascadeHappened = true; t.cascade(val); }
            else super.put(key, val);
        }

        public Object get(Object key) {
            // common case
            if(!(key instanceof String)) return super.get(key);
            if (key.equals("trapee")) return t.trapee;
            if (key.equals("trapname")) return t.name;
            if (key.equals("cascade")) return t.cascade();
            return super.get(key);
        }
    }
}

