// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;
import org.xwt.translators.*;
import org.bouncycastle.util.encoders.Base64;

/** Singleton class that provides all functionality in the xwt.* namespace */
public final class XWT extends JS.Obj {

    public final Res rr;
    public XWT(Res rr) { this.rr = rr; }

    private final JS xwtMath = new XWTMath();
    private final JS xwtString = new XWTString();

    /** each key is a string representing a filename which the user has already given XWT permission to write to */
    private static Hashtable safeFiles = new Hashtable();

    public Object get(Object name) {
        if (name.equals("alt")) return Surface.alt ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("rr")) return rr;
        else if (name.equals("box")) return new Box();
        else if (name.equals("control")) return Surface.control ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("shift")) return Surface.shift ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("clipboard")) return Platform.getClipBoard();
        else if (name.equals("origin")) return Main.origin;
        else if (name.equals("maxdim")) return new Integer(Short.MAX_VALUE);
        else if (name.equals("altKeyName")) return Platform.altKeyName();
        else if (name.equals("screenWidth")) return new Integer(Platform.getScreenWidth());
        else if (name.equals("screenHeight")) return new Integer(Platform.getScreenHeight());
        else if (name.equals("fileSeparator")) return File.separator;
        else if (name.equals("homeDir")) return System.getProperty("user.home");
        else if (name.equals("tempDir")) return System.getProperty("java.io.tempdir");
        else if (name.equals("math")) return xwtMath;
        else if (name.equals("string")) return xwtString;
        else if (name.equals("parseInt")) return xwtString.get("parseInt");
        else if (name.equals("parseFloat")) return xwtString.get("parseFloat");
        else if (name.equals("button")) {
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return new Integer(1);
            else return new Integer(0);
        }
        else return super.get(name);
    }

    public void put(Object name, Object value) {
        if (name.equals("thread") && value != null && value instanceof JS.Callable) ThreadMessage.newthread((JS.Callable)value);
        else if (name.equals("clipboard")) Platform.setClipBoard(value.toString());
        else if (name.equals("proxyAuthorization")) {
            HTTP.Proxy.Authorization.authorization = value.toString();
            HTTP.Proxy.Authorization.waitingForUser.release();
        } else super.put(name, value);
    }

    public Object callMethod(Object method, JS.Array args, boolean checkOnly) {

        if (method.equals("newBrowserWindow")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1 || args.elementAt(0) == null) return null;
            Platform.newBrowserWindow(args.elementAt(0).toString());
            return null;

        } else if (method.equals("clone")) {
            if (checkOnly) return Boolean.TRUE;
            return new XWT((Res)args.elementAt(0));

        } else if (method.equals("watchProgress")) {
            if (checkOnly) return Boolean.TRUE;
            return new Res.ProgressWatcher((Res)args.elementAt(0), (JS.Callable)args.elementAt(1));

        } else if (method.equals("yield")) {
            if (checkOnly) return Boolean.TRUE;
            sleep(0);
            return null;

        } else if (method.equals("load")) {
            if (checkOnly) return Boolean.TRUE;
            return Res.stringToRes(args.elementAt(0).toString());

        } else if (method.equals("println")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1) return null;
            if (Log.on) Log.logJS(this, (args.elementAt(0) == null ? "**null**" : args.elementAt(0).toString()));
            return null;

        } else if (method.equals("date")) {
            if (checkOnly) return Boolean.TRUE;
            Log.log(XWT.class, "date not implemented");
            return null;

        } else if (method.equals("regexp")) {
            if (checkOnly) return Boolean.TRUE;
            return new Regexp(args);

        } else if (method.equals("xmlrpc")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1 || args.elementAt(0) == null) return null;
            return new XMLRPC(args.elementAt(0).toString(), "");

        } else if (method.equals("soap")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() == 1 && args.elementAt(0) != null) return new SOAP(args.elementAt(0).toString(), "", null, null);
            else if (args.length() == 2 && args.elementAt(0) != null && args.elementAt(1) != null)
                return new SOAP(args.elementAt(0).toString(), "", args.elementAt(1).toString(), null);
            else if (args.length() == 3 && args.elementAt(0) != null && args.elementAt(1) != null && args.elementAt(2) != null)
                return new SOAP(args.elementAt(0).toString(), "", args.elementAt(1).toString(), args.elementAt(2).toString());
            else return null;

        } else if (method.equals("sleep")) {
            if (checkOnly) return Boolean.TRUE;
            if (args != null && (args.length() != 1 || args.elementAt(0) == null)) return null;
            int i = args == null ? 0 : Box.stoi(args.elementAt(0).toString());
            sleep(i);
            return null;

        } else if (method.equals("openFile")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1) return null;
            String file = Platform.fileDialog(args.elementAt(0).toString(), false);
            return file == null ? null : Res.stringToRes("file:" + file);

        } else if (method.equals("saveFile")) {
            if (checkOnly) return Boolean.TRUE;
            // FIXME
            /*
            if (args.length() != 2) return null;
            if (!(args.elementAt(1) instanceof ByteStream)) return null;
            String file = args.elementAt(0).toString();
            if (safeFiles.get(Platform.isCaseSensitive() ? file : file.toLowerCase()) == null) {
                file = Platform.fileDialog(file, true);
                if (file == null) return null;
                safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
            }
            try {
                ((ByteStream)args.elementAt(1)).writeTo(new FileOutputStream(file));
                return null;
            } catch (IOException e) {
                if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                if (Log.on) Log.log(ByteStream.class, e);
                throw new JS.Exn("error while writing a ByteStream to a file");
            }
            */

        } else if (method.equals("saveFileAs")) {
            // FIXME
            /*
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 2) return null;
            if (!(args.elementAt(1) instanceof ByteStream)) return null;
            String file = args.elementAt(0).toString();
            file = Platform.fileDialog(file, true);
            if (file == null) return null;
            safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
            try {
                ((ByteStream)args.elementAt(1)).writeTo(new FileOutputStream(file));
                return null;
            } catch (IOException e) {
                if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                if (Log.on) Log.log(ByteStream.class, e);
                throw new JS.Exn("error while writing a ByteStream to a file");
            }
            */

        } else if (method.equals("parseHTML")) {
            if (checkOnly) return Boolean.TRUE;
                if (args == null || args.length() != 1 || args.elementAt(0) == null) return null;
                try {
                    if (args.elementAt(0) instanceof Res) {
                        return HTML.parseReader(new InputStreamReader(((Res)args.elementAt(0)).getInputStream()));
                    } else {
                        return HTML.parseReader(new StringReader(args.elementAt(0).toString()));
                    }
                } catch (IOException e) {
                    if (Log.on) Log.log(HTML.class, "IO Exception while parsing HTML");
                    if (Log.on) Log.log(HTML.class, e);
                    throw new JS.Exn("error while parsing HTML");
                }

        } else if (method.equals("recursivePrintObject")) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1) return null;
            recurse("", "", args.elementAt(0));
            return null;
        }
            
        if (checkOnly) return Boolean.FALSE;
        return null;
    }

    private static void recurse(String indent, String name, Object o) {
        if (!name.equals("")) name += " : ";

        if (o == null) {
            Log.logJS(indent + name + "<null>");

        } else if (o instanceof JS.Array) {
            Log.logJS(indent + name + "<array>");
            JS.Array na = (JS.Array)o;
            for(int i=0; i<na.length(); i++)
                recurse(indent + "  ", i + "", na.elementAt(i));

        } else if (o instanceof JS) {
            Log.logJS(indent + name + "<object>");
            JS s = (JS)o;
            Object[] keys = s.keys();
            for(int i=0; i<keys.length; i++)
                if (keys[i] != null)
                    recurse(indent + "  ", keys[i].toString(),
                            (keys[i] instanceof Integer) ?
                            s.get(((Integer)keys[i])) : s.get(keys[i].toString()));

        } else {
            Log.logJS(indent + name + o);

        }
    }

    public static void sleep(int i) {
        java.lang.Thread thread = java.lang.Thread.currentThread();
        if (!(thread instanceof ThreadMessage)) {
            if (Log.on) Log.log(XWT.class, "cannot sleep() or yield() in the foreground thread");
        } else {
            ThreadMessage mythread = (ThreadMessage)thread;
            mythread.done.release();
            if (i > 0) try { java.lang.Thread.sleep(i); } catch (Exception e) { }
            Message.Q.add(mythread);
            mythread.go.block();
        }
    }
    
    private static class XWTMath extends JS.Obj {
        public XWTMath() {
            JS gs = new JS.GlobalScope();
            put("isNaN",gs.get("isNaN"));
            put("isFinite",gs.get("isFinite"));
            put("NaN",gs.get("NaN"));
            put("Infinity",gs.get("Infinity"));
            setSeal(true);
        }
        public Object get(Object key) {
            Object ret = super.get(key);
            if(ret == null) ret = JS.Math.get(key);
            return ret;
        }
    }
    private static class XWTString extends JS.Obj {
        public XWTString() {
            JS gs = new JS.GlobalScope();
            put("parseInt",gs.get("parseInt"));
            put("parseFloat",gs.get("parseFloat"));
            put("decodeURI",gs.get("decodeURI"));
            put("decodeURIComponent",gs.get("decodeURIComponent"));
            put("encodeURI",gs.get("encodeURI"));
            put("encodeURIComponent",gs.get("encodeURIComponent"));
            put("escape",gs.get("escape"));
            put("unescape",gs.get("unescape"));
            put("fromCharCode",gs.get("stringFromCharCode"));
            setSeal(true);
        }
    }
}
