// Copyright 2003 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt.js;

import org.xwt.util.*;
import java.io.*;

/** a JavaScript function, compiled into bytecode */
class CompiledFunctionImpl extends JS.Callable implements ByteCodes, Tokens {

    // Fields and Accessors ///////////////////////////////////////////////

    /** the number of formal arguments */
    int numFormalArgs = 0;

    /** the source code file that this block was drawn from */
    private String sourceName;
    public String getSourceName() throws JS.Exn { return sourceName; }
    
    /** the line numbers */
    private int[] line = new int[10];

    /** the first line of this script */
    private int firstLine = -1;

    /** the instructions */
    private int[] op = new int[10];

    /** the arguments to the instructions */
    private Object[] arg = new Object[10];

    /** the number of instruction/argument pairs */
    private int size = 0;
    int size() { return size; }

    /** the scope in which this function was declared; by default this function is called in a fresh subscope of the parentScope */
    private JS.Scope parentScope;

    // Constructors ////////////////////////////////////////////////////////

    private CompiledFunctionImpl cloneWithNewParentScope(JS.Scope s) throws IOException {
        CompiledFunctionImpl ret = new JS.CompiledFunction(sourceName, firstLine, null, s);
        // Reuse the same op, arg, line, and size variables for the new "instance" of the function
        // NOTE: Neither *this* function nor the new function should be modified after this call
        ret.op = this.op;
        ret.arg = this.arg;
        ret.line = this.line;
        ret.size = this.size;
        return ret;
    }

    protected CompiledFunctionImpl(String sourceName, int firstLine, Reader sourceCode, JS.Scope parentScope) throws IOException {
        this.sourceName = sourceName;
        this.firstLine = firstLine;
        this.parentScope = parentScope;
        if (sourceCode == null) return;
        Parser p = new Parser(sourceCode, sourceName, firstLine);
        while(true) {
            int s = size();
            p.parseStatement(this, null);
            if (s == size()) break;
        }
        add(-1, LITERAL, null); 
        add(-1, RETURN);
    }
    
    public Object call(JS.Array args) throws JS.Exn { return call(args, new FunctionScope(sourceName, parentScope)); }
    public Object call(JS.Array args, JS.Scope scope) throws JS.Exn {
        JS.Thread cx = JS.Thread.fromJavaThread(java.lang.Thread.currentThread());
        CompiledFunction saved = cx.currentCompiledFunction;
        try {
            cx.currentCompiledFunction = (CompiledFunction)this;
            int size = cx.stack.size();
            cx.stack.push(callMarker);
            cx.stack.push(args);
            eval(scope);
            Object ret = cx.stack.pop();
            if (cx.stack.size() > size)
                // this should never happen
                throw new Error("ERROR: stack grew by " + (cx.stack.size() - size) +
                                " elements during call at " + sourceName + ":" + firstLine);
            return ret;
        } catch(Error e) {
            // Unwind the stack
            while(cx.stack.size() > 0) if(cx.stack.pop() instanceof CallMarker) throw e;
            throw new Error("CallMarker not found on the stack"); // should never happen
        } finally {
            cx.currentCompiledFunction = saved;
        }
    }


    // Adding and Altering Bytecodes ///////////////////////////////////////////////////

    int get(int pos) { return op[pos]; }
    Object getArg(int pos) { return arg[pos]; }
    void set(int pos, int op_, Object arg_) { op[pos] = op_; arg[pos] = arg_; }
    void set(int pos, Object arg_) { arg[pos] = arg_; }
    int pop() { size--; arg[size] = null; return op[size]; }
    void paste(CompiledFunctionImpl other) { for(int i=0; i<other.size; i++) add(other.line[i], other.op[i], other.arg[i]); }
    CompiledFunctionImpl add(int line, int op_) { return add(line, op_, null); }
    CompiledFunctionImpl add(int line, int op_, Object arg_) {
        if (size == op.length - 1) {
            int[] line2 = new int[op.length * 2]; System.arraycopy(this.line, 0, line2, 0, op.length); this.line = line2;
            Object[] arg2 = new Object[op.length * 2]; System.arraycopy(arg, 0, arg2, 0, arg.length); arg = arg2;
            int[] op2 = new int[op.length * 2]; System.arraycopy(op, 0, op2, 0, op.length); op = op2;
        }
        this.line[size] = line;
        op[size] = op_;
        arg[size] = arg_;
        size++;
        return this;
    }
    
    public int getLine(int pc) {
        if(pc < 0 || pc >= size) return -1;
        return line[pc];
    }


    // Invoking the Bytecode ///////////////////////////////////////////////////////
        
    void eval(JS.Scope s) {
        final JS.Thread cx = JS.Thread.fromJavaThread(java.lang.Thread.currentThread());
        final Vec t = cx.stack;
        int pc;
        int lastPC = -1;
        OUTER: for(pc=0; pc<size; pc++) {
        try {
            String label = null;
            cx.pc = lastPC = pc;
            int curOP = op[pc];
            Object curArg = arg[pc];
            if(curOP == FINALLY_DONE) {
                FinallyData fd = (FinallyData) t.pop();
                if(fd == null) continue OUTER; // NOP
                curOP = fd.op;
                curArg = fd.arg;
            }
            switch(curOP) {
            case LITERAL: t.push(arg[pc]); break;
            case OBJECT: t.push(new JS.Obj()); break;
            case ARRAY: t.push(new JS.Array(JS.toNumber(arg[pc]).intValue())); break;
            case DECLARE: s.declare((String)(arg[pc]==null ? t.peek() : arg[pc])); if(arg[pc] != null) t.push(arg[pc]); break;
            case TOPSCOPE: t.push(s); break;
            case JT: if (JS.toBoolean(t.pop())) pc += JS.toNumber(arg[pc]).intValue() - 1; break;
            case JF: if (!JS.toBoolean(t.pop())) pc += JS.toNumber(arg[pc]).intValue() - 1; break;
            case JMP: pc += JS.toNumber(arg[pc]).intValue() - 1; break;
            case POP: t.pop(); break;
            case SWAP: { Object o1 = t.pop(); Object o2 = t.pop(); t.push(o1); t.push(o2); break; }
            case DUP: t.push(t.peek()); break;
            case NEWSCOPE: s = new JS.Scope(s); break;
            case OLDSCOPE: s = s.getParentScope(); break;
            case ASSERT: if (!JS.toBoolean(t.pop())) throw je("assertion failed"); break;
            case BITNOT: t.push(new Long(~JS.toLong(t.pop()))); break;
            case BANG: t.push(new Boolean(!JS.toBoolean(t.pop()))); break;

            case TYPEOF: {
                Object o = t.pop();
                if (o == null) t.push(null);
                else if (o instanceof JS) t.push(((JS)o).typeName());
                else if (o instanceof String) t.push("string");
                else if (o instanceof Number) t.push("number");
                else if (o instanceof Boolean) t.push("boolean");
                else t.push("unknown");
                break;
            }

            case NEWFUNCTION: {
                try {
                    t.push(((CompiledFunctionImpl)arg[pc]).cloneWithNewParentScope(s));
                } catch (IOException e) {
                    throw new Error("this should never happen");
                }
                break;
            }

            case PUSHKEYS: {
                Object o = t.peek();
                Object[] keys = ((JS)o).keys();
                JS.Array a = new JS.Array();
                a.setSize(keys.length);
                for(int j=0; j<keys.length; j++) a.setElementAt(keys[j], j);
                t.push(a);
                break;
            }

            case LABEL: break;
            case LOOP: {
                t.push(new LoopMarker(pc, pc > 0 && op[pc - 1] == LABEL ? (String)arg[pc - 1] : (String)null,s));
                t.push(Boolean.TRUE);
                break;
            }

            case BREAK:
            case CONTINUE:
                while(t.size() > 0) {
                    Object o = t.pop();
                    if (o instanceof CallMarker) ee("break or continue not within a loop");
                    if (o instanceof TryMarker) {
                        if(((TryMarker)o).finallyLoc < 0) continue; // no finally block, keep going
                        t.push(new FinallyData(curOP, curArg));
                        s = ((TryMarker)o).scope;
                        pc = ((TryMarker)o).finallyLoc - 1;
                        continue OUTER;
                    }
                    if (o instanceof LoopMarker) {
                        if (curArg == null || curArg.equals(((LoopMarker)o).label)) {
                            int loopInstructionLocation = ((LoopMarker)o).location;
                            int endOfLoop = ((Integer)arg[loopInstructionLocation]).intValue() + loopInstructionLocation;
                            s = ((LoopMarker)o).scope;
                            if (curOP == CONTINUE) { t.push(o); t.push(Boolean.FALSE); }
                            pc = curOP == BREAK ? endOfLoop - 1 : loopInstructionLocation;
                            continue OUTER;
                        }
                    }
                }
                throw new Error("CONTINUE/BREAK invoked but couldn't find a LoopMarker at " + sourceName + ":" + getLine(pc));

            case TRY: {
                int[] jmps = (int[]) arg[pc];
                // jmps[0] is how far away the catch block is, jmps[1] is how far away the finally block is
                // each can be < 0 if the specified block does not exist
                t.push(new TryMarker(jmps[0] < 0 ? -1 : pc + jmps[0], jmps[1] < 0 ? -1 : pc + jmps[1],s));
                break;
            }

            case RETURN: {
                Object retval = t.pop();
                while(t.size() > 0) {
                    Object o = t.pop();
                    if (o instanceof TryMarker) {
                        if(((TryMarker)o).finallyLoc < 0) continue;
                        t.push(retval); 
                        t.push(new FinallyData(RETURN));
                        s = ((TryMarker)o).scope;
                        pc = ((TryMarker)o).finallyLoc - 1;
                        continue OUTER;
                    }
                    if (o instanceof CallMarker) {
                        t.push(retval);
                        return;
                    }
                }
                throw new Error("error: RETURN invoked but couldn't find a CallMarker!");
            }

            case PUT: {
                Object val = t.pop();
                Object key = t.pop();
                Object target = t.peek();
                if (target == null)
                    throw je("tried to put a value to the " + key + " property on the null value");
                if (!(target instanceof JS))
                    throw je("tried to put a value to the " + key + " property on a " + target.getClass().getName());
                if (key == null)
                    throw je("tried to assign \"" + (val==null?"(null)":val.toString()) + "\" to the null key");
                ((JS)target).put(key, val);
                t.push(val);
                break;
            }

            case GET:
            case GET_PRESERVE: {
                Object o, v;
                if (op[pc] == GET) {
                    v = arg[pc] == null ? t.pop() : arg[pc];
                    o = t.pop();
                } else {
                    v = t.pop();
                    o = t.peek();
                    t.push(v);
                }
                Object ret = null;
                if (o == null) throw je("tried to get property \"" + v + "\" from the null value");
                if (v == null) throw je("tried to get the null key from " + o);
                if (o instanceof String || o instanceof Number || o instanceof Boolean)
                    ret = Internal.getFromPrimitive(o,v);
                else if (o instanceof JS)
                    ret = ((JS)o).get(v);
                else 
                    throw je("tried to get property " + v + " from a " + o.getClass().getName());
                t.push(ret);
                break;
            }
            
            case CALLMETHOD:
            case CALL:
            {
                JS.Array arguments = new JS.Array();
                int numArgs = JS.toNumber(arg[pc]).intValue();
                arguments.setSize(numArgs);
                for(int j=numArgs - 1; j >= 0; j--) arguments.setElementAt(t.pop(), j);
                Object o = t.pop();
                if(o == null) throw je("attempted to call null");
                Object ret;
                if(op[pc] == CALLMETHOD) {
                    Object method = o;
                    o = t.pop();
                    if(o instanceof String || o instanceof Number || o instanceof Boolean)
                        ret = Internal.callMethodOnPrimitive(o,method,arguments);
                    else if(o instanceof JS)
                        ret = ((JS)o).callMethod(method,arguments,false);
                    else
                        throw new JS.Exn("Tried to call a method on an object that isn't a JS object");
                } else {                   
                    ret = ((JS.Callable)o).call(arguments);
                }
                t.push(ret);
                break;
            }
            // fall through if exception was thrown
            case THROW: {
                Object o = t.pop();
                if(o instanceof JS.Exn) throw (JS.Exn)o;
                throw new JS.Exn(o);
            }

            case INC: case DEC: {
                boolean isPrefix = JS.toBoolean(arg[pc]);
                Object key = t.pop();
                JS obj = (JS)t.pop();
                Number num = JS.toNumber(obj.get(key));
                Number val = new Double(op[pc] == INC ? num.doubleValue() + 1.0 : num.doubleValue() - 1.0);
                obj.put(key, val);
                t.push(isPrefix ? val : num);
                break;
            }
            
            case ASSIGN_SUB: case ASSIGN_ADD: {
                Object val = t.pop();
                Object old = t.pop();
                Object key = t.pop();
                Object obj = t.peek();
                if (val instanceof CompiledFunction) {
                    if (obj instanceof JS.Scope) {
                        JS.Scope parent = (JS.Scope)obj;
                        while(parent.getParentScope() != null) parent = parent.getParentScope();
                        if (parent instanceof org.xwt.Box) {
                            if (curOP == ASSIGN_ADD) {
                                ((org.xwt.Box)parent).addTrap(key, val);
                            } else {
                                ((org.xwt.Box)parent).delTrap(key, val);
                            }
                            // skip over the "normal" implementation of +=/-=
                            pc += ((Integer)arg[pc]).intValue() - 1;
                            break;
                        }
                    }
                }
                // use the "normal" implementation
                t.push(key);
                t.push(old);
                t.push(arg);
                break;
            }

            case ADD: {
                int count = ((Number)arg[pc]).intValue();
                if(count < 2) throw new Error("this should never happen");
                if(count == 2) {
                    // common case
                    Object right = t.pop();
                    Object left = t.pop();
                    if(left instanceof String || right instanceof String) t.push(JS.toString(left).concat(JS.toString(right)));
                    else t.push(new Double(JS.toDouble(left) + JS.toDouble(right)));
                } else {
                    Object[] args = new Object[count];
                    while(--count >= 0) args[count] = t.pop();
                    if(args[0] instanceof String) {
                        StringBuffer sb = new StringBuffer(64);
                        for(int i=0;i<args.length;i++) sb.append(JS.toString(args[i]));
                        t.push(sb.toString());
                    } else {
                        int numStrings = 0;
                        for(int i=0;i<args.length;i++) if(args[i] instanceof String) numStrings++;
                        if(numStrings == 0) {
                            double d = 0.0;
                            for(int i=0;i<args.length;i++) d += JS.toDouble(args[i]);
                            t.push(new Double(d));
                        } else {
                            double d=0.0;
                            int i=0;
                            do {
                                d += JS.toDouble(args[i++]);
                            } while(!(args[i] instanceof String));
                            StringBuffer sb = new StringBuffer(64);
                            sb.append(JS.toString(new Double(d)));
                            while(i < args.length) sb.append(JS.toString(args[i++]));
                            t.push(sb.toString());
                        }
                    }
                }
                break;
            }

            default: {
                Object right = t.pop();
                Object left = t.pop();
                switch(op[pc]) {
                        
                case BITOR: t.push(new Long(JS.toLong(left) | JS.toLong(right))); break;
                case BITXOR: t.push(new Long(JS.toLong(left) ^ JS.toLong(right))); break;
                case BITAND: t.push(new Long(JS.toLong(left) & JS.toLong(right))); break;

                case SUB: t.push(new Double(JS.toDouble(left) - JS.toDouble(right))); break;
                case MUL: t.push(new Double(JS.toDouble(left) * JS.toDouble(right))); break;
                case DIV: t.push(new Double(JS.toDouble(left) / JS.toDouble(right))); break;
                case MOD: t.push(new Double(JS.toDouble(left) % JS.toDouble(right))); break;
                        
                case LSH: t.push(new Long(JS.toLong(left) << JS.toLong(right))); break;
                case RSH: t.push(new Long(JS.toLong(left) >> JS.toLong(right))); break;
                case URSH: t.push(new Long(JS.toLong(left) >>> JS.toLong(right))); break;
                        
                case LT: case LE: case GT: case GE: {
                    if (left == null) left = new Integer(0);
                    if (right == null) right = new Integer(0);
                    int result = 0;
                    if (left instanceof String || right instanceof String) {
                        result = left.toString().compareTo(right.toString());
                    } else {
                        result = (int)java.lang.Math.ceil(JS.toDouble(left) - JS.toDouble(right));
                    }
                    t.push(new Boolean((op[pc] == LT && result < 0) || (op[pc] == LE && result <= 0) ||
                                       (op[pc] == GT && result > 0) || (op[pc] == GE && result >= 0)));
                    break;
                }
                    
                case EQ:
                case NE: {
                    Object l = left;
                    Object r = right;
                    boolean ret;
                    if (l == null) { Object tmp = r; r = l; l = tmp; }
                    if (l == null && r == null) ret = true;
                    else if (r == null) ret = false; // l != null, so its false
                    else if (l instanceof Boolean) ret = new Boolean(JS.toBoolean(r)).equals(l);
                    else if (l instanceof Number) ret = JS.toNumber(r).doubleValue() == JS.toNumber(l).doubleValue();
                    else if (l instanceof String) ret = r != null && l.equals(r.toString());
                    else ret = l.equals(r);
                    t.push(new Boolean(op[pc] == EQ ? ret : !ret)); break;
                }

                default: throw new Error("unknown opcode " + op[pc]);
                } }
            }
        } catch(JS.Exn e) {
            while(t.size() > 0) {
                Object o = t.pop();
                if (o instanceof CatchMarker || o instanceof TryMarker) {
                    boolean inCatch = o instanceof CatchMarker;
                    if(inCatch) {
                        o = t.pop();
                        if(((TryMarker)o).finallyLoc < 0) continue; // no finally block, keep going
                    }
                    if(!inCatch && ((TryMarker)o).catchLoc >= 0) {
                        // run the catch block, this will implicitly run the finally block, if it exists
                        t.push(o);
                        t.push(catchMarker);
                        t.push(e.getObject());
                        s = ((TryMarker)o).scope;
                        pc = ((TryMarker)o).catchLoc - 1;
                        continue OUTER;
                    } else {
                        t.push(e);
                        t.push(new FinallyData(THROW));
                        s = ((TryMarker)o).scope;
                        pc = ((TryMarker)o).finallyLoc - 1;
                        continue OUTER;
                    }
                }
                // no handler found within this func
                if(o instanceof CallMarker) throw e;
            }
            throw new Error("couldn't find a Try or Call Marker!");
        } // end try/catch
        } // end for
        // this should never happen, we will ALWAYS have a RETURN at the end of the func
        throw new Error("Just fell out of CompiledFunction::eval() loop. Last PC was " + lastPC);
    }

    // Debugging //////////////////////////////////////////////////////////////////////

    public String toString() {
        StringBuffer sb = new StringBuffer(1024);
        sb.append("\n" + sourceName + ": " + firstLine + "\n");
        for (int i=0; i < size; i++) {
            sb.append(i);
            sb.append(": ");
            if (op[i] < 0)
                sb.append(bytecodeToString[-op[i]]);
            else
                sb.append(codeToString[op[i]]);
            sb.append(" ");
            sb.append(arg[i] == null ? "(no arg)" : arg[i]);
            if((op[i] == JF || op[i] == JT || op[i] == JMP) && arg[i] != null && arg[i] instanceof Number) {
                sb.append(" jump to ").append(i+((Number) arg[i]).intValue());
            } else  if(op[i] == TRY) {
                int[] jmps = (int[]) arg[i];
                sb.append(" catch: ").append(jmps[0] < 0 ? "No catch block" : ""+(i+jmps[0]));
                sb.append(" finally: ").append(jmps[1] < 0 ? "No finally block" : ""+(i+jmps[1]));
            }
            sb.append("\n");
        }
        return sb.toString();
    } 

    // Exception Stuff ////////////////////////////////////////////////////////////////

    static class EvaluatorException extends RuntimeException { public EvaluatorException(String s) { super(s); } }
    EvaluatorException ee(String s) { throw new EvaluatorException(sourceName + ":" + JS.Thread.currentJSThread().getLine() + " " + s); }
    JS.Exn je(String s) { throw new JS.Exn(sourceName + ":" + JS.Thread.currentJSThread().getLine() + " " + s); }


    // FunctionScope /////////////////////////////////////////////////////////////////

    private static class FunctionScope extends JS.Scope {
        String sourceName;
        public FunctionScope(String sourceName, Scope parentScope) { super(parentScope); this.sourceName = sourceName; }
        public String getSourceName() { return sourceName; }
    }


    // Markers //////////////////////////////////////////////////////////////////////

    public static class CallMarker { public CallMarker() { } }
    private static CallMarker callMarker = new CallMarker();
    
    public static class CatchMarker { public CatchMarker() { } }
    private static CatchMarker catchMarker = new CatchMarker();
    
    public static class LoopMarker {
        public int location;
        public String label;
        public JS.Scope scope;
        public LoopMarker(int location, String label, JS.Scope scope) {
            this.location = location;
            this.label = label;
            this.scope = scope;
        }
    }
    public static class TryMarker {
        public int catchLoc;
        public int finallyLoc;
        public JS.Scope scope;
        public TryMarker(int catchLoc, int finallyLoc, JS.Scope scope) {
            this.catchLoc = catchLoc;
            this.finallyLoc = finallyLoc;
            this.scope = scope;
        }
    }
    public static class FinallyData {
        public int op;
        public Object arg;
        public FinallyData(int op, Object arg) { this.op = op; this.arg = arg; }
        public FinallyData(int op) { this(op,null); }
    }
}

/** this class exists solely to work around a GCJ bug */
abstract class JSCallable extends JS.Callable {
        public abstract Object call(JS.Array args) throws JS.Exn;
}
